# pylint: disable=C0325,C0103,C0111
from __future__ import print_function
import os
import sys
import logging
import docopt
import six
from murano_client.atomicconfigparser import atomicconfigparser as CfgParser
from murano_client.atomicconfigparser import NoOptionError, NoSectionError
if six.PY2:
    prompt = raw_input
elif six.PY3:
    prompt = input

LOG = logging.getLogger('MURANO_CLIENT.cfg')

class Command(object):
    """System wide INI file editor and viewer.

  usage:
    cfg <file> [get <section> <option>]
    cfg <file> set <section> [<option> <value> -U]

  options:
    -h --help                   Show this screen.
    get                         Print a section or option.
    set                         Update or create a new section or option.
    <file>                      The name of the INI file.
    <section>                   The name of the INI file section.
    <option>                    The name of the INI option.
    <value>                     The name of the INI option value.
    --unset -U                  Remove the section or option.

    """
    Name = 'cfg'
    def __init__(self, command_args, global_args):
        """
        Initialize the commands.
        :param command_args: arguments of the command
        :param global_args: arguments of the program
        """
        self.args = docopt.docopt(self.__doc__, argv=command_args)
        global_args.update({k:v for k, v in self.args.items() if v})
        self.global_args = global_args

        if self.global_args.get('--debug'):
            LOG.setLevel(eval('logging.'+self.global_args.get('--debug')))

    def execute(self):

        the_file = self.global_args.get('<file>')

        if not os.path.exists(the_file):
            print("No config file exists at: {}".format(the_file))
            sys.exit(-1)

        parser = CfgParser(allow_no_value=True)
        parser.read(the_file)

        if self.global_args.get('get'):
            to_print = ''
            try:
                if self.global_args.get('<section>') and self.global_args.get('<option>'):
                    to_print = parser.get(
                        self.global_args.get('<section>'),
                        self.global_args.get('<option>')
                    )
                elif self.global_args.get('<section>'):
                    for item in parser.items(self.global_args.get('<section>')):
                        if item[1]:
                            to_print += item[0] + ' = ' + item[1]
                        else:
                            to_print += item[0]
                        to_print += '\n'
                else:
                    to_print = open(the_file, 'r').read()
                print(to_print, end='')
            except NoOptionError as exc:
                print(exc.message)
                sys.exit(-1)
            except NoSectionError as exc:
                print(exc.message)
                sys.exit(-1)
        elif self.global_args.get('set'):
            with open(the_file, 'r') as cfg:
                parser.readfp(cfg)
                if self.global_args.get('<section>'):
                    if not self.global_args.get('--unset'):
                        if not parser.has_section(self.global_args.get('<section>')):
                            parser.add_section(self.global_args.get('<section>'))
                        if self.global_args.get('<option>'):
                            parser.set(
                                self.global_args.get('<section>'),
                                self.global_args.get('<option>'),
                                self.global_args.get('<value>')
                            )
                    else:
                        if not self.global_args.get('<option>'):
                            # unset used only with section
                            parser.remove_section(self.global_args.get('<section>'))
                        else:
                            parser.remove_option(
                                self.global_args.get('<section>'),
                                self.global_args.get('<option>')
                            )
                    parser.write(the_file)
        else:
            print(open(the_file, 'r').read())



"""
A subclass of the Configuration file parser that support atomic updates.

That is, if a write to a configuration file on a POSIX compatible file system
and the write is interrupted (e.g. by a power failure), the file will either
have the original contents or the updated contents.  The file will not be
corrupted.
"""

import tempfile as _tempfile
import os as _os
import fnmatch as _fn
from sys import stderr
import logging
import six
if six.PY2:
    import ConfigParser as _cfg
    from ConfigParser import NoOptionError, NoSectionError
elif six.PY3:
    import configparser as _cfg
    from configparser import NoOptionError, NoSectionError

LOG = logging.getLogger(name='MURANO_CLIENT.atomicconfigparser')

class atomicconfigparser(_cfg.RawConfigParser):
    """
        The atomicconfigparser is for ensuring atomic writes to files
        when on jffs2 and ubifs filesystems.
    """
    def __init__(self, **kwargs):
        _cfg.RawConfigParser.__init__(self, **kwargs)
        # preserving the case sensitivity
        self.optionxform = str

    def write(self, filename):
        """Write an .ini-format representation of the configuration state atomically."""
        
        # The configuration state is written to a temporary file in the same
        # directory as filename.  That assures us that the files will be on the
        # same file system.  Once the state is succesively written, the
        # temporary file will be renamed to filename.

        # Get the path for the desired filename.  The temporary file will be
        # created in this location.
        path = _os.path.dirname(_os.path.abspath(filename))

        # Create a temp file in the directory that will not be deleted when
        # it's closed.
        with _tempfile.NamedTemporaryFile(
            'w',
            prefix='atomic-',
            suffix='-cfg',
            dir=path,
            delete=False) as tmpf:
            # Call base class write routine to create the file
            _cfg.RawConfigParser.write(self, tmpf)

            # Flush the data to the temporary file
            tmpf.flush()

            # Sync the file contents to disk
            _os.fsync(tmpf.fileno())

            # Save the name  so that we can rename the file later
            tempname = tmpf.name

        # Rename the file from its temporary name to the actual name.  This
        # action, on a POSIX system, is atomic.
        _os.rename(tempname, filename)

        # Open the directory containing the temporary file and sync it as well
        # to make sure the rename operation has made it to storage.
        dirfd = _os.open(path, _os.O_DIRECTORY)
        _os.fsync(dirfd)
        _os.close(dirfd)

        # Clean up any old temporary configuration files that may exist.  They
        # will match the filename pattern 'atomic-*-cfg'.  This can happen
        # if a write is interrupted by a power cycle.
        for _file in _os.listdir(path):
            if _fn.fnmatch(_file, 'atomic-*-cfg'):
                try:
                    _os.unlink(_file)
                except OSError as err:
                    stderr.write("atomicconfigparser: could not unlink {0}: {1}\n"
                                 .format(_file, err), stderr)
                    stderr.write("atomicconfigparser: does file exist? {}\n"
                                 .format(_os.path.exists(_file)))
                except Exception as err: # pylint: disable=W0703
                    print("atomicconfigparser: unknown error: {0}".format(err), stderr)




# pylint: disable=C0325,C0103
from __future__ import print_function
from murano_client.commands.http import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args

class ExoCommand(ExositeConnection):
    """Use the HTTP Device API to activate client described by Product
and Device IDs. Prints the CIK to STDOUT if successful. If -f option is
used, the Device config file is used.

  usage:
    timestamp [options]

  options:
    -h --help                   Show this screen.

    """

    def execute(self):
        from murano_client.http import MuranoHTTP
        from murano_client.ini import Device

        if self.global_args.get('--file'):
            api = Device(self.global_args.get('--file'))
        else:
            api = MuranoHTTP(**build_api_opts_from_docopt_args(self.global_args))
        print(api.exosite_timestamp().text, end='') # suppress newline



# pylint: disable=C0103,C0301,C0325,W1202

import os
import ssl
import logging
import time
import six
from threading import Timer
from paho.mqtt.client import Client as PahoSNIMqttClient
from murano_client import BaseMuranoClient, __version__
if six.PY2:
    from urlparse import urlparse
elif six.PY3:
    from urllib.parse import urlparse

PahoReturnCodes = {
    0: "success, connection accepted",
    1: "connection refused, bad protocol",
    2: "refused, client-id error",
    3: "refused, service unavailable",
    4: "refused, bad username or password",
    5: "refused, not authorized"
}

LOG = logging.getLogger(name='MURANO_CLIENT.MuranoMQTT')


class MqttMsg(object):
    """
        This class is a wrapper class for MQTT
        messages. It is likely that he paho client
        will be used basicaly forever, but in the
        event that the protocol library changes,
        wrapping the message object will help
        us change the underlying library without
        breaking changes propagating upstream.
    """
    def __init__(self, msg):
        self.dup = msg.dup
        self.info = msg.info
        self.mid = msg.mid
        self.payload = msg.payload
        self.qos = msg.qos
        self.retain = msg.retain
        self.state = msg.state
        self.timestamp = msg.timestamp
        self.topic = msg.topic

class MqttPublishHandler(object):
    """
        ['__class__', '__delattr__', '__doc__', '__format__', '__getattribute__', '__getitem__',
        '__hash__', '__init__', '__iter__', '__module__', '__new__', '__next__', '__reduce__',
        '__reduce_ex__', '__repr__', '__setattr__', '__sizeof__', '__slots__', '__str__',
        '__subclasshook__', '_condition', '_iterpos', '_published', '_set_as_published',
        'is_published', 'mid', 'next', 'rc', 'wait_for_publish']
    """
    def __init__(self, msg_info):
        self.rc = msg_info.rc
        self.mid = msg_info.mid
        self.is_published = msg_info.is_published
        self.wait_for_publish = msg_info.wait_for_publish
        self.next = msg_info.next

class MuranoMQTT(PahoSNIMqttClient, BaseMuranoClient, object):
    def __init__(self, **kwargs):
        """
        :param pid: Specify the Murano Product ID. This will set the Vendor and the Model member variables to the same.

        :type  vendor: string
        """
        BaseMuranoClient.__init__(self, **kwargs)
        PahoSNIMqttClient.__init__(self, client_id="")
        _ = LOG.setLevel(getattr(logging, kwargs.get('debug'))) if kwargs.get('debug') else None

        LOG.debug("kwargs: {}".format(kwargs))
        tls_set_args = {'cert_reqs': ssl.CERT_OPTIONAL}
        if self.murano_cacert():
            tls_set_args.update(ca_certs=self.murano_cacert())

        if self.using_tls():
            tls_set_args.update({
                'certfile': self.certfile(),
                'keyfile': self.pkeyfile(),
            })
        LOG.info("tls_set_args: {}".format(tls_set_args))
        self.tls_set(**tls_set_args)
        # TODO: this is unacceptable: https://github.com/eclipse/paho.mqtt.python#tls_insecure_set
        self.tls_insecure_set(kwargs.get('tls_insecure_set') or False)
        if kwargs.get("password"):
            uuid = self.murano_id()
        else:
            uuid = ""
        pw = self.murano_token()
        LOG.info("setting username/password: id: {}, token: {}"
                 .format(uuid, pw))
        self.username_pw_set(uuid, pw)
        if not kwargs.get('murano_port'):
            LOG.debug("no murano_port specified. default to 8883")
            self.set_murano_port(8883)
        else:
            LOG.debug("murano_port overridden in kwargs: {}".format(kwargs.get('murano_port')))
            self.set_murano_port(kwargs.get('murano_port'))
        LOG.debug("mqtt host: {}, port: {}".format(
            self.murano_host(), self.murano_port()))

        self.on_log = self.default_on_log
        self.on_connect = self.default_on_connect
        self.on_disconnect = self.default_on_disconnect
        self.on_message = self.default_on_message
        self.on_publish = self.default_on_publish

        self.time_connected = 0.0
        self._connected = False
        self.is_connected = lambda: self._connected
        self.duration_connected = lambda: time.time() - self.time_connected
        self.put_messages_in_this_queue = kwargs.get('queue_inbound_messages')
        self.disconnect_after_seconds = kwargs.get('disconnect_after_seconds')
        if self.disconnect_after_seconds:
            self.disconnect_after_seconds = float(self.disconnect_after_seconds)
        self.always_reconnect = True

    def __str__(self):
        return 'id: {} host: {} port: {} auth: {}'.format(
            self.murano_id(), self.murano_host(), self.murano_port(), self.using_tls())

    def default_on_log(self, client, userdata, level, buf):
        """ Default mqtt logger."""
        LOG.log(level, "client: {} userdata: {} buf: {} mid: {}"
                .format(client, userdata, buf, buf[buf.find('(')+len('Mid: ')+1:buf.find(')')]))

    def default_on_message(self, client, userdata, msg):
        """ Default on_message function for tunable logging. """
        LOG.info("dup: {} info: {} mid: {} qos: {} retain: {} state: {} timestamp: {} topic: {}"
                 .format(msg.dup,
                         msg.info,
                         msg.mid,
                         msg.qos,
                         msg.retain,
                         msg.state,
                         msg.timestamp,
                         msg.topic))
        LOG.debug("client: {} userdata: {} payload: {}"
                  .format(str(client), userdata, msg.payload))
        if self.put_messages_in_this_queue:
            LOG.info("Putting message in user-defined queue: {}".format(msg))
            self.put_messages_in_this_queue.put(msg)

    def default_on_publish(self, client, userdata, result):
        """ Default on_publish method for logging. """
        LOG.info("client: {} userdata: {} result (mid): {}"
                 .format(client, userdata, result))

    def default_on_connect(self, client, userdata, flags, rc):
        """ Default on_connect method for tracking connection status and logging. """
        LOG.info("client: {} userdata: {} flags: {} rc: {} reason: {}"
                 .format(client, userdata, flags, rc, PahoReturnCodes[rc]))
        self.time_connected = time.time()
        self._connected = True
        if self.disconnect_after_seconds:
            self.always_reconnect = False
            disconnect_timer = Timer(float(self.disconnect_after_seconds), self.disconnect)
            disconnect_timer.start()
            LOG.info("disconnecting in {} seconds".format(self.disconnect_after_seconds))

    def default_on_disconnect(self, client, userdata, rc):
        """ Default on_disconnect method for tracking connection status and logging. """
        LOG.debug("client: {} userdata: {} rc: {} reason: {}"
                  .format(client, userdata, rc, PahoReturnCodes[rc]))
        LOG.info("time connected: {}".format(time.time()-self.time_connected))
        self.time_connected = 0.0
        self._connected = False
        if rc != 0 and self.always_reconnect:
            LOG.warning("ExositeMQTT default disconnection handler: {}, {}"\
                        .format(rc, PahoReturnCodes[rc]))
            self.connect(urlparse(self.murano_host()).netloc, self.murano_port())

    def mqtt_activate(self):
        default_on_message = self.on_message

        def on_message(self, userdata, msg):
            LOG.debug("Activation succeeded!")
            self.set_murano_token(msg.payload.decode())
            LOG.info(self.murano_token())
            self.username_pw_set("", self.murano_token())
            LOG.debug("activate_on_message: {} :: {} :: {}"
                      .format(self, userdata, msg.payload))

        if not self.using_tls():
            LOG.debug("Using token or user/pw method of provisioning.")
            self.on_message = on_message
            LOG.debug("publishing activate message...")
            rc, mid = self.publish("$provision/" + self.murano_id(), "", qos=1)
            LOG.debug("activation publish rc: {}, mid: {}, reason: {}".format(rc, mid, PahoReturnCodes[rc]))
            LOG.info("Waiting for activation...")
        else:
            LOG.debug("Using TLS Client Cert for provisioning.")

        start_time = time.time()
        while not self.is_activated():
            time.sleep(0.1)
            if time.time() - start_time >= 1.0:
                LOG.debug(".")
            if self.disconnect_after_seconds:
                if time.time() - start_time >= self.disconnect_after_seconds:
                    LOG.warning("Aborting activation due to disconnect_after_seconds: {}"
                                .format(self.disconnect_after_seconds))
                    break
        if self.is_activated():
            LOG.debug("Activated. Took {} seconds.".format(time.time()-start_time))
        else:
            LOG.debug("Activation attempt was unsuccessful after {} seconds"
                      .format(time.time()-start_time))
        self.on_message = default_on_message

    def start(self, activate_only=False):
        LOG.info("starting...")
        try:
            self.connect(urlparse(self.murano_host()).netloc, self.murano_port())
            self.loop_start()
            start_time = time.time()
            while not self.is_connected():
                time.sleep(0.1)
                if time.time() - start_time >= 1.0:
                    LOG.debug(".")
            if not self.is_activated():
                self.mqtt_activate()
        except ssl.SSLError as exc:
            LOG.error(dir(exc))
            LOG.error(exc.errno)
            LOG.error(exc.reason)
            LOG.error(exc.strerror)
        LOG.info("Started. Time taken to connect: {}.".format(time.time()-start_time))
        if activate_only:
            self.loop_stop()
            self.disconnect()




import unittest
from murano_client.client import MuranoClient, MuranoClientException

class TestMuranoClient(unittest.TestCase):
    def test_no_kwargs(self):
        with self.assertRaises(MuranoClientException):
            MuranoClient()
    def test_no_watchlist(self):
        with self.assertRaises(MuranoClientException):
            MuranoClient(
                murano_host='mqtt://f5330e5s8cho0000.m2.exosite.io/')
    def test_good_mqtt_url(self):
        client = MuranoClient(
            watchlist=['config_io'],
            murano_host='mqtt://f5330e5s8cho0000.m2.exosite.io/'
        )
        self.assertEquals('mqtt', client.outbound_protocol)
    def test_good_http_url(self):
        client = MuranoClient(
            watchlist=['config_io'],
            murano_host='https://f5330e5s8cho0000.m2.exosite.io/'
        )
        self.assertEquals('https', client.outbound_protocol)

def main():
    unittest.main()

if __name__ == "__main__":
    main()


# pylint: disable=C0325,C0103,C0111
import logging
import time
from murano_client.commands.mqtt import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args

LOG = logging.getLogger('GDC.MQTT.sub')

class ExoCommand(ExositeConnection):
    """Subscribe to Murano device using the MQTT protocol.

  usage:
    subscribe [options]

  options:
    -h --help                       Show this screen.
    --disconnect-after <seconds>    Only subscribe for <seconds> seconds.

    """
    Name = 'subscribe'
    def execute(self):
        from murano_client.ini import Device
        from murano_client.mqtt import MuranoMQTT

        if self.global_args.get('--file'):
            mqtt_client = Device(self.global_args.get('--file'))
        else:
            mqtt_client = MuranoMQTT(**build_api_opts_from_docopt_args(self.global_args))

        def on_message(client, userdata, msg):
            _, resource, timestamp = msg.topic.split('/')[0:3]
            print("{}.{}={}".format(resource, timestamp, msg.payload.decode()))
        mqtt_client.on_message = on_message

        mqtt_client.start()

        try:
            start = time.time()
            mqtt_client.loop_start()
            while True:
                if mqtt_client.disconnect_after_seconds:
                    if time.time() - start >= mqtt_client.disconnect_after_seconds:
                        break
                time.sleep(0.1)
        except KeyboardInterrupt:
            pass



# pylint: disable=C0325,C0103,W0212
from murano_client.commands.http import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args
import docopt

class ExoCommand(ExositeConnection):
    """Record <value> to <alias> at time <timestamp>.

  usage:
    content [<subcommand> <content_name>] [options]

  subcommands:
    list
    info <content_name>
    download <content_name>

  options:
    -h --help                   Show this screen.

    """

    def execute(self):
        from murano_client.http import MuranoHTTP
        from murano_client.ini import Device

        if self.global_args.get('--file'):
            api = Device(self.global_args.get('--file'))
        else:
            api = MuranoHTTP(**build_api_opts_from_docopt_args(self.global_args))

        if self.args.get('<subcommand>') == 'list':
            resp = api._http_listContent()
            if resp.status_code == 200:
                print(resp.text)
            else:
                print("[{}] {}".format(resp.status_code, resp.text))
            return
        if not self.args.get('<content_name>'):
            raise docopt.DocoptExit("subcommand: {!r} requires <content_name>."
                                    .format(self.args.get('<subcommand>')))
        if self.args.get('<subcommand>') == 'info':
            resp = api._http_getContentInfo(self.args.get('<content_name>'))
            print("[{}] {}".format(resp.status_code, resp.text))
            return
        elif self.args.get('<subcommand>') == 'download':
            resp = api._http_getContent(self.args.get('<content_name>'))
            if resp.status_code == 200:
                with open(self.args.get('<content_name>'), 'wb+') as __f:
                    for chunk in resp.iter_content(chunk_size=1024):
                        if chunk:
                            __f.write(chunk)
            print("[{}] {}".format(resp.status_code, self.args.get('<content_name>')))



"""
    This file has a garbage version string in it 
    intentionally so that the build.sh script at
    the top of this repository can create the
    version string according to calver.org 
    convention.
"""
__version__ = '0.0.0'
__version__ = '18.8.10'



""" His module"""

# pylint: disable=I0011,W0312,R0903,C1001,R0201,W0232,C0301,W1202

import os
import logging
from murano_client.client import MuranoClient
# from murano_client.http import MuranoHTTP
# from murano_client.mqtt import MuranoMQTT
from murano_client.atomicconfigparser import atomicconfigparser as IniParser

LOG = logging.getLogger(name='MURANO_CLIENT.ini')

class ExoDeviceException(Exception):
    """ Custom exception class for validating
        user input and improper usage of ExoTLS class.
    """
    def __init__(self, message):
        super(ExoDeviceException, self).__init__(message)

class Device(MuranoClient, object):
    """
        TODO
    """
    def __init__(self, ini_file):
        self._ini_parser = IniParser(allow_no_value=True)
        self._ini_parser.optionxform = str
        self._ini_file = ini_file
        if not os.path.exists(self._ini_file):
            raise ExoDeviceException("Provided config file {!r} does not exist.".format(self._ini_file))
        parent_config = self.get_config_from_ini_file()
        LOG.debug("using config: {}".format(parent_config))
        _ = LOG.setLevel(getattr(logging, parent_config.get('debug'))) if parent_config.get('debug') else None
        MuranoClient.__init__(self, **parent_config)

    def set_murano_host(self, host):
        """ TODO """
        super(Device, self).set_murano_host(host)
        self.dump_to_ini_file('device', 'murano_host', host)

    def set_murano_port(self, port):
        """ TODO """
        super(Device, self).set_murano_port(port)
        self.dump_to_ini_file('device', 'murano_port', port)

    def set_murano_token(self, token):
        """ TODO """
        LOG.debug("TOKEN SET: {}".format(token))
        super(Device, self).set_murano_token(token)
        if token:
            self.dump_to_ini_file('device', 'murano_token', token)

    def set_murano_id(self, uuid):
        """ TODO """
        super(Device, self).set_murano_id(uuid)
        self.dump_to_ini_file('device', 'murano_id', uuid)

    def ini_file(self):
        """ Protected member getter. """
        return self._ini_file

    def get_config_from_ini_file(self):
        """
            Reads INI config files and converts them to a dictionary.
        """
        config = {}
        self._ini_parser.read(self.ini_file())
        sections = self._ini_parser.sections()
        if 'device' not in sections:
            raise ExoDeviceException("No 'device' section found in {}".format(self.ini_file()))
        for section in sections:
            for option in self._ini_parser.options(section):
                # anything in the 'device' section should be added to the
                # config dictionary as top-level keys that will be inputs
                # to the ExositeAPI class as kwargs. any other config sections
                # will be passed with the section as the top-level key to
                # the config settings.
                if 'device' == section:
                    # To accomodate allow_no_value=True, options with no value return
                    # None, which evaluates to False. Since options with no value are
                    # used as booleans that (when present) mean True, so if a get() on
                    # a given option returns None, set to True instead
                    value = self._ini_parser.get(section, option)
                    # if option == 'debug':
                    #     LOG.critical(str((option, value)))
                    #     value = getattr(logging, value, 'INFO')
                    if option == 'murano_port':
                        value = int(value)
                    if option == 'watchlist':
                        value = value.split(',')
                    config[option] = True if value is None else value
                else:
                    # see comment above for the boolean magic explanation
                    config[section] = {} if not config.get(section) else config.get(section)
                    value = self._ini_parser.get(section, option)
                    config[section][option] = True if value is None else value
        return config

    def update_device_from_cfg(self):
        """ Updates Device instance with values from cfg. """
        LOG.debug("Updating member variables from config file: {!r}"
                  .format(self.ini_file()))
        self._ini_parser.read(self.ini_file())

        self.set_murano_token(
            self._ini_parser.get(
                'device', 'murano_token'
            )
        )
        self.set_murano_id(
            self._ini_parser.get(
                'device', 'uuid'
            )
        )

        if self._ini_parser.has_option('device', 'murano_host'):
            self.set_murano_host(
                self._ini_parser.get(
                    'device', 'murano_host'
                )
            )
        if self._ini_parser.has_option('device', 'http_timeout'):
            self.set_http_timeout(
                self._ini_parser.getfloat('device', 'http_timeout')
            )


    def dump_to_ini_file(self, section, option, value):
        """ Utility to dump configuration data to the instance cfg file. """
        if os.path.exists(self.ini_file()):
            with open(self.ini_file(), 'r') as cfg:
                self._ini_parser.readfp(cfg)
                if self._ini_parser.has_section(section):
                    self._ini_parser.set(section, option, value)
                else:
                    LOG.debug("Unable to dump: '{!r}' '{!r}' '{!r}'"
                              .format(section, option, value))
                    return # don't bother writing it, we didn't update anything.
            self._ini_parser.write(self.ini_file())
        else:
            LOG.debug("Can't find configuration file: {!r}".format(self.ini_file()))

    def activate_device(self):
        """ If using TOKEN auth, use the appropriate protocol to retrieve a token and
        save it to the INI file. If using TLS Provisioning, don't do really anything."""
        self.client_activate()
        if self.is_activated():
            if self.using_tls():
                LOG.info("Using TLS Provisioning. No INI file updates needed.")
            else:
                LOG.info("Updating Murano Token in INI file: {}".format(self.ini_file()))
                self.dump_to_ini_file('device', 'murano_token', self.murano_token())




# pylint: disable=W1202,R0902,W0312
"""
    This module contains the MuranoClient and MuranoTLS
    classes which are used to build protocol clients.

    E.g.
        class MqttClient(MuranoClient):
            ...

        qtclient = MqttClient(host='kjli97yjuhmk00000.m2.exosite.io')
        qtclient.start()
"""
from __future__ import print_function
import os
import sys
import json
import logging
import pkg_resources
from murano_client.logger import LOG
from .__version__ import __version__
if sys.version_info[0] == 2: # six.PY2:
    from urlparse import urlparse
elif sys.version_info[0] >=3:
    from urllib.parse import urlparse

class ExoTLSException(Exception):
    """ Custom exception class for validating
        user input and improper usage of ExoTLS class.
    """
    pass

class MuranoTLS(object):
    """ Container class for filesystem paths to
    TLS Client Cert Auth and CA files. """


    def __init__(self, **kwargs):
        self._murano_cacert = kwargs.get('murano_cacert') or \
            pkg_resources.resource_filename('murano_client', 'Murano_Root_CA.cer')
        self._certfile = kwargs.get('certfile')
        self._pkeyfile = kwargs.get('pkeyfile')
        self.murano_cacert = lambda: self._murano_cacert
        self.certfile = lambda: self._certfile
        self.pkeyfile = lambda: self._pkeyfile
        if self.certfile() and not os.path.exists(self.certfile()):
            raise ExoTLSException("Certfile {!r} not found!"
                                  .format(self.certfile()))
        if self.pkeyfile() and not os.path.exists(self.pkeyfile()):
            raise ExoTLSException("Keyfile {!r} not found!"
                                  .format(self.pkeyfile()))

    def using_tls(self):
        """ Returns a boolean to determine whether or not
        the client is using TLS Client Cert Auth. """
        using_tls = None not in (self.certfile(), self.pkeyfile())
        LOG.debug("Using TLS Provisioning? {}".format(using_tls))
        return using_tls

class MuranoClientException(Exception):
    """ Custom exception class for validating
        user input and improper usage of
        BaseMuranoClient class.
    """
    pass

class BaseMuranoClient(MuranoTLS, object):
    """
        A base class for building protocol clients.

        A Murano Client can include but aren't necessarily
        limited to the following attributes and methods.

         - A client supports the following provisioning attributes:
            1. Token
            2. Username/Password
            3. TLS Client Certificate

         - A client supports the following provisioning methods:
            1. onep:v1 activate - provide data to provisioning
               endpoints/topics to get CIK/Token in a response
            2. username/password - use
            3. TLS Client Cert

    """
    def __init__(self, **kwargs):
        # initialize tls vars
        MuranoTLS.__init__(self, **kwargs)
        # state variables
        _ = LOG.setLevel(getattr(logging, kwargs.get('debug'))) if kwargs.get('debug') else None
        self.__version__ = __version__
        self._activated = False
        self._online = False

        # provisioning and connection parameters
        self._murano_id = kwargs.get('murano_id')
        self._murano_host = kwargs.get('murano_host')
        self._murano_port = kwargs.get('murano_port') or 443
        self.set_murano_token(kwargs.get('murano_token') or None)
        if self.using_tls():
            self.set_activated(True)
            LOG.info("Using TLS, setting activated...")
            self._auth_header = lambda: {}
        else:
            self._auth_header = lambda: {'X-Exosite-CIK' : self.murano_token()}

        self.murano_host = lambda: self._murano_host
        self.murano_port = lambda: self._murano_port
        self.murano_id = lambda: self._murano_id
        self.murano_token = lambda: self._murano_token

        if urlparse(self.murano_host()).scheme == 'mqtt':
            self.set_murano_port(8883)

    def set_murano_host(self, host):
        """ TODO """
        self._murano_host = host

    def set_murano_port(self, port):
        """ TODO. """
        self._murano_port = port

    def set_murano_token(self, token):
        """ Return current murano auth token. """
        self._murano_token = token
        if token:
            self.set_activated(True)

    def set_murano_id(self, uuid):
        """ Return current murano auth token. """
        self._murano_id = uuid

    def set_activated(self, boolean):
        """ Set whether or not the client is
        activated. """
        self._activated = boolean

    def is_activated(self):
        """ Return a boolean to determine
        whether or not the client is activated. """
        if self.using_tls():
            return True
        return self._activated





"""
    Global namespace for gmq.commands.
"""
# pylint: disable=C0103,C0325

import docopt

def build_api_opts_from_docopt_args(args):
    api_opts = {}
    _ = api_opts.update(
        {'debug': args.get('--debug')})
    _ = api_opts.update(
        {'murano_id': args.get('--uuid')})
    _ = api_opts.update(
        {'murano_token': args.get('--token')})
    _ = api_opts.update(
        {'http_timeout': args.get('--timeout'),
         'disconnect_after_seconds': args.get('--disconnect-after')})
    _ = api_opts.update(
        {'murano_host': args.get('--host')})
    _ = api_opts.update(
        {'murano_port': args.get('--port')})
    _ = api_opts.update(
        {'certfile': args.get('--cert'),
         'pkeyfile': args.get('--pkey')})
    _ = api_opts.update({'murano_cacert': args.get('--cacert')})
    return api_opts

# TODO: The Command class in both http and mqtt are 99% the same.
#       The classes should be broken out, generalized and abstracted
#       here so they may be either subclassed or monkey patched in
#       their respective command __init__ files.


# pylint: disable=C0325,C0103,C0111,W1202
from __future__ import print_function
import threading
import logging
import time
from sys import stdin
from murano_client.commands.mqtt import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args

LOG = logging.getLogger('GDC.MQTT.pubsub')

class ExoCommand(ExositeConnection):
    """Pipe streaming data to a Murano resource while simultaneously
    subscribing to the devices resources.

  usage:
    pubsub <topic> [options]

  options:
    -h --help                       Show this screen.
    <topic>                         Mqtt topic to publish to Murano resource.
    --disconnect-after <seconds>    Only subscribe for <seconds> seconds.

    """
    Name = 'pubsub'
    def execute(self):
        from murano_client.ini import Device
        from murano_client.mqtt import MuranoMQTT

        if self.global_args.get('--file'):
            mqtt_client = Device(self.global_args.get('--file'))
        else:
            mqtt_client = MuranoMQTT(**build_api_opts_from_docopt_args(self.global_args))

        mqtt_client.start()

        Qos = 1
        if self.global_args.get('--qos'):
            Qos = int(self.global_args.get('--qos'))

        topic = self.args.get('<topic>')
        LOG.debug("Publishing to topic: {}".format(topic))
        def pub():
            def read_stdin():
                readline = stdin.readline()
                while readline:
                    yield readline
                    readline = stdin.readline()
            rough_count = 0
            for line in read_stdin():
                payload = line.strip()
                LOG.debug("PAYLOAD: {}".format(payload))
                mqtt_client.publish(
                    topic,
                    payload,
                    qos=Qos
                )
                rough_count += 1
                LOG.debug("connected: {} - {}"
                          .format(mqtt_client.connected(), mqtt_client.duration_connected()))
        pubthread = threading.Thread(target=pub, name='PublishThread')
        pubthread.setDaemon(True)

        def on_message(client, userdata, msg):
            """ Override default on_message function. """
            _, resource, timestamp = msg.topic.split('/')[0:3]
            print("{}.{}={}".format(resource, timestamp, msg.payload))
        mqtt_client.on_message = on_message
        try:
            pubthread.start()
            mqtt_client.loop_start()
            start = time.time()
            while True:

                if mqtt_client.disconnect_after_seconds:
                    if time.time() - start >= mqtt_client.disconnect_after_seconds:
                        break
                time.sleep(0.1)
            exit(0)
        except KeyboardInterrupt:
            pass




# pylint: disable=C0325,C0103
from murano_client.commands.http import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args

class ExoCommand(ExositeConnection):
    """Read the most recent data from a resource <alias>.

  usage:
    read <alias> [<alias>]... [options]

  options:
    -h --help                   Show this screen.
    <alias>                     A Murano Product resource.

    """
    Name = 'read'
    def execute(self):
        from murano_client.http import MuranoHTTP
        from murano_client.ini import Device

        if self.global_args.get('--file'):
            api = Device(self.global_args.get('--file'))
        else:
            api = MuranoHTTP(**build_api_opts_from_docopt_args(self.global_args))

        aliases = self.args.get('<alias>')
        resp_handler = api.http_read(
            aliases if isinstance(aliases, list) else [aliases])

        print("[{}] {}".format(resp_handler.code, resp_handler.body if 204 != resp_handler.code else "No Content"))



# pylint: disable=C0325,C0103
from murano_client.commands.http import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args

class ExoCommand(ExositeConnection):
    """Use the HTTP Device API to activate client described by Product
and Device IDs. Prints the CIK to STDOUT if successful. If -f option is
used, the Device config file is used.

  usage:
    activate [options]

  options:
    -h --help                   Show this screen.

    """
    Name = 'activate'
    def execute(self):
        from murano_client.http import MuranoHTTP
        from murano_client.ini import Device

        if self.global_args.get('--file'):
            d = Device(self.global_args.get('--file'))
            d.activate_device()
            if d.is_activated():
                with open(self.global_args.get('--file'), 'r') as f:
                    print(f.read())
            else:
                print("Activation attempt not successful.")
        else:
            api = MuranoHTTP(**build_api_opts_from_docopt_args(self.global_args))
            resp_handler = api._http_activate()
            if resp_handler.activated:
                api.set_murano_token(resp_handler.body)
                print(api.murano_token())
            else:
                print("[{}] {}".format(resp_handler.code, resp_handler.body if resp_handler.code != 204 else "No Content"))




# pylint: disable=C0325,C0103,C0111
from murano_client.commands.mqtt import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args

class ExoCommand(ExositeConnection):
    """Activate a Murano client (retreive the cik/token) via the MQTT protocol.

  usage:
    activate [options]

  options:
    -h --help                   Show this screen.
    -i --pid <pid>              The Product ID.
    -u --uuid <uuid>            Specify the username (mqtt only) or serial 
                                number (id) of the connecting client. During
                                provisioning or when using TLS Client Cert
                                auth, this is left blank/not used.
    -f --file <file>            All operations will use <file> as the Device state file. This file
                                is compatible with Device class objects. Will store <cik> and
                                provision as the Murano client described in its options. This
                                option has special support for configuring the GWE config file
                                by using 'gwe' as the <file> argument.

    """

    def execute(self):
        from murano_client.mqtt import MuranoMQTT
        from murano_client.ini import Device

        if self.global_args.get('--file'):
            mqtt_client = Device(self.global_args.get('--file'))
        else:
            mqtt_client = MuranoMQTT(**build_api_opts_from_docopt_args(self.global_args))

        mqtt_client.start(activate_only=True)
        print(mqtt_client.murano_token())









# pylint: disable=W1202
"""
    Namespace for mqtt subcommand of gdc.commands.
"""

import sys
import logging
import six
import docopt
import pkg_resources
import importlib

LOG = logging.getLogger(name='GDC.MQTT')

class Command(object): # ExositeConnection
    """Use methods of the MQTT(S) Device API.

  usage:
    mqtt <command> [<args>...] [options]

  commands:
    activate        Retreive CIKs/Tokens for devices
    publish         Publish data to Murano resources.
    subcribe        Subscribe to Murano resources.
    pubsub          Publish and subscribe to Murano resources.

  args:
    <command>                   The HTTP(S) Device API method name.
    <args>                      Args for subcommands.

  options:
    -h --help                       Show this screen.
    -N --no-ssl                     If provided, SSL/TLS will not be used in HTTP(S) comms.
    -P --no-product-domain          If provided, project domains (i.e. your project ID).
    -i --pid <id>                   The Product ID.
    -t --timeout <secs>             If provided, <secs> is used as HTTP Timeout (in seconds).
                                    Not implemented in Mqtt.
    -H --host <host>                If provided, <host> is used as the server hostname for
                                    HTTP requests.
    -d --debug <lvl>                Turn on verbose debug output. Also logs curl commands.
    -u --uuid <uuid>                Specify the id of the connecting client (mqtt only).
    -f --file <file>                All operations will use <file> as the Device state file. This file
                                    is compatible with Device class objects. Will store <cik> and
                                    provision as the Murano client described in its options. This
                                    option has special support for configuring the GWE config file
                                    by using 'gwe' as the <file> argument.
    -C --cert <cert>                Use <cert> for Murano Provisioning of client described in
                                    the certificate subject.
    -K --pkey <pkey>                Use <pkey> in TLS communication with Murano.
    -c --cik <cik>                  A CIK (Client Interface Key), Token or password.
    -E --cacert <cacert>            Set the CA cert for the connection to the Murano Product
                                    (default:located in wheel - murano_client/Murano_Root_CA.cer).
                                    This can be changed if using a PKI system.
    -Q --qos <qos>                  Determine the Quality of Service for published
                                    messages (default=1). Possible values:

                                        0 - Message sent At Most Once (supported)
                                        1 - Message sent At Least Once (supported)
                                        2 - Message sent Exactly Once (not supported)
    --disconnect-after <seconds>    Only subscribe or pubsub for <seconds> seconds.

    """
    Name = 'mqtt'
    def __init__(self, command_args, global_args):
        """
        Initialize the commands.
        :param command_args: arguments of the command
        :param global_args: arguments of the program
        """
        self.args = docopt.docopt(self.__doc__, argv=command_args)
        global_args.update({k:v for k, v in self.args.items() if v})
        self.global_args = global_args

        if self.global_args.get('--debug'):
            LOG.setLevel(getattr(logging, self.global_args.get('--debug')))
            LOG.debug('global_args: {}'.format(self.global_args))

    def execute(self, **exc_args):
        """Execute the commands"""
        # Retrieve the command to execute.
        command_name = self.args.pop('<command>')
        # Retrieve the command arguments.
        command_args = self.args.pop('<args>')

        try:
            if pkg_resources.resource_exists('murano_client.commands.mqtt', command_name+'.py'):
                LOG.debug('loading from module: {}'
                          .format(pkg_resources.resource_filename(
                              'murano_client.commands.mqtt', command_name)))
                the_module = importlib.import_module('murano_client.commands.mqtt.' + command_name)
            command_class = getattr(the_module, 'ExoCommand')

        except ImportError as exc:
            LOG.debug("{}: cannot find command {!r}: {}".format(self.Name, command_name, exc))
            raise docopt.DocoptExit()
        except AttributeError:
            LOG.debug('{}: unknown command: {}'.format(self.Name, command_name))
            raise docopt.DocoptExit()

        # Create an instance of the command and execute it
        command_class(command_args, self.global_args).execute()



# pylint: disable=I0011,W0312,C0301,C0103,W0123,W1202
r"""Gateway Device Client (GDC) Command-Line interface.

GDC is a light-weight cli for common tasks using the device-client library.

Usage:
  gdc [options] [<command>] [<args>...]

Commands:
    cfg             Interface for viewing and modifying gdc configuration settings.
    http            Use the HTTP(S) Device API from the command line.
    mqtt            Start and mqtt client. Only available on Python3.x environments.

Options:
    -h --help                   Show this screen.
    -v --version                Print the current version of device-client.
    -f --file <file>            All operations will use <file> as the Device state file. This file
                                is compatible with Device class objects. Will store <cik> and
                                provision as the Murano client described in its options. This
                                option has special support for configuring the GWE config file
                                by using 'gwe' as the <file> argument.
    -t --timeout <secs>         If provided, <secs> is used as HTTP Timeout (in seconds).
                                Not implemented in Mqtt.
    -H --host <host>            If provided, <host> is used as the server hostname for
                                HTTP requests.
    -P --port <port>            If provided, <port> is used as the TCP port used for
                                communicating with Murano (default=443).
    -d --debug <lvl>            Tune the debug output (DEBUG|INFO|WARNING|ERROR|CRITICAL).
                                Logs curl commands at DEBUG.
    -u --uuid <uuid>            The id or username of a Murano device..
    -C --cert <cert>            Use <cert> for Murano Provisioning of client described in
                                the certificate subject.
    -K --pkey <pkey>            Use <pkey> in TLS communication with Murano.
    -E --cacert <cacert>        Set the CA cert for the connection to the Murano Product
                                (default:/etc/exosite/gdc/Murano_Root_CA.cer). This can be
                                changed if using a PKI system.
    -k --token <token>          The authorization token or password of a Murano device.
    <command>                   The gdc subcommand name.
    <args>                      Supported arguments for <command>.

"""
from __future__ import print_function
import imp
import os
import sys
import pkgutil
import logging
import importlib
import pkg_resources
from docopt import docopt
from docopt import DocoptExit
import murano_client.commands
from murano_client import __version__ as VERSION

LOG = logging.getLogger(name='GDC')
LOG_FORMAT = "%(asctime)s-%(levelname)s-%(name)s:%(funcName)s:%(lineno)d ::> %(message)s"
FORMATTER = logging.Formatter(LOG_FORMAT)
streamh = logging.StreamHandler(sys.stdout)
streamh.setFormatter(FORMATTER)
LOG.addHandler(streamh)
LOG.propagate = False


def main():
    global_args = docopt(__doc__, version=VERSION, options_first=True)
    # print(global_args)
    if global_args.get('--version'):
        print(VERSION)
        return

    if global_args.get('--debug'):
        LOG.setLevel(getattr(logging, global_args.get('--debug')))

    if global_args.get('<command>'):
        # Retrieve the command to execute.
        command_name = global_args.pop('<command>')
        # Retrieve the command arguments.
        argv = global_args.pop('<args>')
        if argv is None:
            argv = {}

        # Retrieve the module from the 'commands' package.
        try:
            if pkg_resources.resource_exists('murano_client.commands', command_name+'.py'):
                LOG.debug('loading from module: {}'.format(pkg_resources.resource_filename('murano_client.commands', command_name)))
                the_module = importlib.import_module('murano_client.commands.' + command_name)
            elif pkg_resources.resource_exists('murano_client.commands.' + command_name, '__init__.py'):
                LOG.debug('loading from package: {}'.format(pkg_resources.resource_filename('murano_client.commands', command_name)))
                the_module = importlib.import_module('murano_client.commands.' + command_name)

            command_class = getattr(the_module, 'Command')
        except ImportError as exc:
            raise DocoptExit("Cannot find command {!r}: {}"
                             .format(command_name, exc))

        global_args['--port'] = int(global_args.get('--port')) if global_args.get('--port') else None
        # Create an instance of the command.
        command = command_class(argv, global_args)

        # Execute the command.
        command.execute()
        return

    raise DocoptExit("provide an option or subcommand\n")

if __name__ == '__main__':
    main()




""" Installer for Murano Client protocol layer. """
# pylint: disable=I0011,W0312,C0410
import os
from murano_client.__version__ import __version__
from setuptools import setup, find_packages

DOCS_URL = 'github.com/exosite/lib_murano_client_python'

REQUIREMENTS = [
    'docopt>=0.6.2',
    'requests>=2.13.0',
    'paho-mqtt',
    'six'
]

def read(fname):
    """ Primarily used to open README file. """
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

try:
    README = read('README.rst')
except:
    README = ''

setup(
    name="murano_client",
    version=__version__,
    author="Exosite LLC",
    author_email="support@exosite.com",
    description="""Murano Device Client is the primary library for communicating with the Exosite Murano Platform.""",
    license="Apache 2.0",
    keywords="murano exosite iot iiot client gateway",
    url="https://github.com/exosite/lib_murano_client_python",
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'gdc = murano_client.gdc:main',
        ]
    },
    install_requires=REQUIREMENTS,
    long_description=README,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Operating System :: POSIX :: Linux",
        "Topic :: System :: Operating System Kernels :: Linux",
        "Topic :: Software Development :: Embedded Systems",
        "License :: OSI Approved :: Apache Software License",
    ],
    include_package_data=True,
)



"""This module contains all HTTP Device API and HTTP Provision Management
methods for IoT gatewayss.
"""

# pylint: disable=I0011,W0312,R0903,C1001,R0201,W0232,C0301,W1202

import sys
from murano_client import __version__, BaseMuranoClient
if sys.version_info[0] < 3:
    import httplib
    from urllib import urlencode, quote_plus, unquote_plus, unquote
else:
    import http.client as httplib
    from urllib.parse import urlencode, quote_plus, unquote_plus, unquote

import logging
import sys
import json
import requests
import time
import os
# from murano_client.logger import requests_log

LOG = logging.getLogger(name='MURANO_CLIENT.MuranoHTTP')


class Http_ActivationCodes:
    """ Enum class for http response codes from Activation attempts.

        `Referenced <http://docs.exosite.com/http/#activate>`_ on 04/16/2015

        ::

            HTTP/1.1 200 OK
            Date: <date>
            Server: <server>
            Connection: Keep-Alive
            Content-Length: <length>
            Content-Type: text/plain; charset=utf-8

        On success, the response is the TOKEN.

        Response may also be:

        * HTTP/1.1 404 Not Found if the client described by <vendor>, <model>, <sn> is not found on the system.
        * HTTP/1.1 409 Conflict if the serial number is not enabled for activation.

        See HTTP Responses for a full list of responses

    """
    Timeout = 0
    OK = 200
    DeviceNotin1P = 403
    NotFound = 404
    NotEnabled = 409

class Http_ReadWriteCodes:
    """ Enum class for http response codes from dataport Read, Write and ReadWrite attempts.

        Copied from docs.exosite.com/data/#http 04/16/2015

        Typical HTTP response codes include:
        Code    Response        Description
        200     OK              Successful request, returning requested values
        204     No Content      Successful request, nothing will be returned
        4xx     Client Error    There was an error* with the request by the client
        401     Unauthorized    No or invalid TOKEN
        5xx     Server Error    There was an error with the request on the server

        * Note: aliases that are not found are not considered errors in the request.

        See the documentation for read, and write and Hybrid write/read for details.
    """
    Timeout = 0
    OK = 200
    NoContent = 204
    Unauthorized = 401
    NotModified = 304
    Conflict = 409
    ClientErrors = [400]+[c for c in range(402, 409)]+[c for c in range(410, 500)]
    ServerErrors = [c for c in range(500, 600)]

class ActivationHandler(object):
    """ Class to handle Device() activations.

        Input: Requests_Response() object generated from Exosite Activation Request.

        Object contains 'code', 'body', and 'activated' member variables.

        * If successfully Activated, member variables will contain:
            * code = 200
            * body = '<token>'
            * activated = True
        * If successfully Activated, but got a BAD TOKEN, member variables will contain:
            * code = 200
            * body = '<BAD TOKEN>'
            * activated = False
        * Otherwise:
            * code = <http response code>
            * body = 'an informative message'
            * activated = False

        Example usage:

        .. code-block:: python

            act_resp = exo.handlers.ActivationHandler(
                exo.Exo().exosite_activate(
                    uuid,
                    vendor,
                    msg.model
                )
            )

            # determine whether or not provisioning/activation worked
            if(act_resp.activated):
                my_token = act_resp.body
            else:
                print("Activation attempt FAILED: {!r}".format(act_resp.body))

    """
    def __init__(self, response):
        """
            :param response: The response object that the handler intends to parse/process.

            :type  response: Requests_Response
        """
        self.response = response
        self.code = response.status_code
        self.body = response.text
        self.activated = False

        if self.code == Http_ActivationCodes.OK:
            if len(self.body) == 40:
                self.activated = True
        elif self.code == Http_ActivationCodes.NotFound:
            self.body = "Client <id> is not found on the system."
        elif self.code == Http_ActivationCodes.DeviceNotin1P:
            self.body = "Received 403 from activation attempt. Device Not in Murano"
        elif self.code == Http_ActivationCodes.NotEnabled:
            self.body = "Received 409 from activation attempt. Device Not Enabled"
        elif self.code == Http_ActivationCodes.Timeout:
            self.body = "Request timed out"
        else:
            self.body = "code: {" + str(self.code) + "} :: Something went wrong."
    def __repr__(self):
        return 'code: {!r}, body: {!r}, success: {!r}'.format(
            self.code, self.body, self.activated)

class WriteHandler(object):
    """
TODO
    """
    def __init__(self, response):
        self.response = response
        self.code = response.status_code
        self.body = response.text
        self.online = True
        self.success = False
        self.authorized = True

        if self.code == Http_ReadWriteCodes.OK or self.code == Http_ReadWriteCodes.NoContent:
            self.body = "No content"
            self.success = True
        elif self.code == Http_ReadWriteCodes.Unauthorized:
            self.body = "No or invalid TOKEN."
            self.authorized = False
        elif self.code in Http_ReadWriteCodes.ClientErrors:
            self.body = "Response code: {" + str(self.code) + "} :: \
There was an error* with the request by the client.  Error Message: " + self.body
        elif self.code in Http_ReadWriteCodes.ServerErrors:
            self.body = "Response code: {" + str(self.code) + "} :: \
There was an error with the request on the server."
        elif self.code == Http_ReadWriteCodes.Timeout:
            self.online = False
    def __repr__(self):
        return 'code: {!r}, body: {!r}, success: {!r}, authorized: {!r}'.format(
            self.code, self.body, self.success, self.authorized)

class RecordHandler(object):
    """
TODO
    """
    def __init__(self, response):
        self.response = response
        self.code = response.status_code
        self.body = response.text
        self.online = True
        self.success = False
        self.authorized = True
        self.conflict = False

        if self.code == Http_ReadWriteCodes.OK or self.code == Http_ReadWriteCodes.NoContent:
            self.body = "No content"
            self.success = True
        elif self.code == Http_ReadWriteCodes.Unauthorized:
            self.body = "No or invalid TOKEN."
            self.authorized = False
        elif self.code == Http_ReadWriteCodes.Conflict:
            self.conflict = True
        elif self.code in Http_ReadWriteCodes.ClientErrors:
            self.body = "Response code: {" + str(self.code) + "} :: \
There was an error* with the request by the client."
        elif self.code in Http_ReadWriteCodes.ServerErrors:
            self.body = "Response code: {" + str(self.code) + "} :: \
There was an error with the request on the server."
        elif self.code == Http_ReadWriteCodes.Timeout:
            self.online = False
    def __repr__(self):
        return('code: {!r}, body: {!r}, success: {!r}, authorized: {!r}, online: {!r}, conflict: {!r}, response: {}'
               .format(self.code, self.body, self.success, self.authorized, self.online, self.conflict, self.response))

    def get_conflict_timestamps(self):
        """
            Returns a list of timestamps that were marked as
            conflicting by Exosite from the body of the response.
        """
        # get rid of % encoding, throw away the dataport/alias name
        # by using index-1 of the '=' split
        data = unquote_plus(self.body).split('=')[1]
        return [float(x) for x in data.split(',')]

class ReadHandler(object):
    """
Hard-wired to urllib.unquote(url).decode('utf-8') all
responses.
    """
    def __init__(self, response):
        self.response = response
        self.code = response.status_code

        if sys.version_info[0] < 3:
            self.body = unquote_plus(response.text).decode('utf-8')
        else:
            self.body = unquote_plus(response.text)
        self.online = True
        self.success = False
        self.authorized = True
        # self.last_modified = None

        if self.code == Http_ReadWriteCodes.OK or self.code == Http_ReadWriteCodes.NoContent:
            self.success = True
        elif self.code == Http_ReadWriteCodes.Unauthorized:
            self.body = "No or invalid TOKEN."
            self.authorized = False
        elif self.code == Http_ReadWriteCodes.NotModified:
            self.body = "Not modified"
        elif self.code in Http_ReadWriteCodes.ClientErrors:
            self.body = "There was an error* with the request by the client."
        elif self.code in Http_ReadWriteCodes.ServerErrors:
            self.body = "There was an error with the request on the server."
        elif self.code == Http_ReadWriteCodes.Timeout:
            self.online = False

        """
            TODO: the last-modified header is human-readable. Need to parse this.
            Exception in thread HTTPWatchThread-config_io:
            Traceback (most recent call last):
              File "/usr/local/Cellar/python@2/2.7.15/Frameworks/Python.framework/Versions/2.7/lib/python2.7/threading.py", line 801, in __bootstrap_inner
                self.run()
              File "/usr/local/lib/python2.7/site-packages/murano_client-1.0.development-py2.7.egg/murano_client/client.py", line 116, in run
                modify_ts=since)
              File "/usr/local/lib/python2.7/site-packages/murano_client-1.0.development-py2.7.egg/murano_client/http.py", line 421, in http_long_poll
                return ReadHandler(self.send_get(url, headers))
              File "/usr/local/lib/python2.7/site-packages/murano_client-1.0.development-py2.7.egg/murano_client/http.py", line 256, in __init__
                self.last_modified = int(self.response.headers.get('Last-Modified'))
            ValueError: invalid literal for int() with base 10: 'Fri, 15 Jun 2018 19:04:13 GMT'
        """

        # if self.response.headers.get('Last-Modified'):
        #     self.last_modified = int(self.response.headers.get('Last-Modified'))
    def __repr__(self):
        return 'code: {}, body: {}, success: {}, authorized: {}'.format(
            self.code, self.body, self.success, self.authorized)

class ReadWriteHandler(object):
    """
Hard-wired to urllib.unquote(url).decode('utf-8') all
responses.
    """
    def __init__(self, response):
        self.response = response
        self.code = response.status_code
        self.body = unquote(response.text).decode('utf-8')
        self.online = True
        self.success = False
        self.authorized = True

        if self.code == Http_ReadWriteCodes.OK or self.code == Http_ReadWriteCodes.NoContent:
            self.success = True
        elif self.code == Http_ReadWriteCodes.Unauthorized:
            self.body = "No or invalid TOKEN."
            self.authorized = False
        elif self.code in Http_ReadWriteCodes.ClientErrors:
            self.body = "There was an error* with the request by the client."
        elif self.code in Http_ReadWriteCodes.ServerErrors:
            self.body = "There was an error with the request on the server."
        elif self.code == Http_ReadWriteCodes.Timeout:
            self.online = False
    def __repr__(self):
        return 'code: {!r}, body: {!r}, success: {!r}, authorized: {!r}'.format(
            self.code, self.body, self.success, self.authorized)


class MuranoHTTP(BaseMuranoClient, object):
    """Instances of the ExositeAPI class contain all methods and members needed to
    interface with the Exosite HTTP Data API. ExositeAPI objects are very versatile and are
    used in a multitude of Gateway Engine components like ``gwe`` and ``gmq``. It is also the
    base-class for the Device class which is a INI-style configuration file based class.

    """
    def __init__(self, **kwargs):
        """
            TODO
        """
        _ = LOG.setLevel(getattr(logging, kwargs.get('debug'))) if kwargs.get('debug') else None
        LOG.debug("init kwargs: {}".format(kwargs))
        BaseMuranoClient.__init__(self, **kwargs)
        self._user_agent = '{}-{}'.format(self.__class__.__name__, __version__)
        self._timeout_secs = 60.0 if not kwargs.get('http_timeout') else float(kwargs.get('http_timeout'))
        self._form_encode = kwargs.get('form_encode')
        self._custom_auth = kwargs.get('custom_auth')
        self._session = None
        self._raise = kwargs.get('raise_exceptions')

        self._activate_url = lambda: self.murano_host() + 'provision/activate'
        self._content_url = lambda: self.murano_host() + 'provision/download?'
        self._stack_alias_url = lambda: self.murano_host() + 'onep:v1/stack/alias'
        self._stack_record_url = lambda: self.murano_host() + 'onep:v1/stack/record'
        self._http_accept = 'application/x-www-form-urlencoded; charset=utf-8'
        self._http_content_type = 'application/x-www-form-urlencoded; charset=utf-8'

        if self._custom_auth and not isinstance(self._custom_auth, dict):
            assert isinstance(self._custom_auth, dict), \
                "Custom headers must be dictionaries if overridding default (None)."

        if self.using_tls():
            self.set_activated(True)
            self._auth_header = lambda: {}
        else:
            self._auth_header = lambda: {'X-Exosite-CIK' : self.murano_token()} \
                if not isinstance(self._custom_auth, dict) \
                else self._custom_auth

    def __str__(self):
        """ So we can print self representation by calling str(self). """
        return str({
            'version': __version__,
            'murano_host': self.murano_host(),
            'murano_id': self.murano_id()
        })

    def no_form_encode(self):
        """ Turns off the usage of SSL in HTTP API."""
        self._form_encode = False
    def form_encode(self):
        """ Turns on the usage of SSL in HTTP API."""
        self._form_encode = True
    def set_http_debug(self, enable):
        """Sets httplib debug level when 'enable' is set to True. """
        if enable:
            httplib.HTTPConnection.debuglevel = 1
        else:
            httplib.HTTPConnection.debuglevel = 0
    def set_http_timeout(self, timeout):
        self._timeout_secs = timeout
    def timeout(self):
        """ returns :code:`self._timeout_secs`
        """
        return self._timeout_secs
    def url_decode(self, obj):
        """ Helper function to urllib.url_decode objects read from dataports. """
        return unquote_plus(obj)

    def close_session(self):
        self._session.close()

    def _http_listContent(self):
        """ Method for retrieving a list of Content area contents. """
        headers = {'User-Agent' : self._user_agent}
        headers.update(self._auth_header())
        LOG.debug('headers: {}'.format(headers))
        return self.send_get(self._content_url(), headers)

    def _http_getContent(self, content_id, stream=None):
        """ Method to retrieve specific content from Content area. """
        headers = {'User-Agent' : self._user_agent}
        headers.update(self._auth_header())
        url = self._content_url() + '&id={}'.format(quote_plus(content_id))
        return self.send_get(url, headers, stream=stream)

    def _http_getContentInfo(self, content_id):
        """ Method to retrieve information on Content area content. """
        headers = {'User-Agent' : self._user_agent}
        headers.update(self._auth_header())
        url = self._content_url() + 'id={}&info=true'.format(quote_plus(content_id))

        return self.send_get(url, headers)

    def _http_read(self, read_list):
        '''
            read_list: List of data sources to read from
        '''
        headers = {
            'User-Agent' : self._user_agent,
            'Accept' : self._http_accept
        }
        headers.update(self._auth_header())
        url = self._stack_alias_url()
        if read_list:
            url = url + '?'
            for alias in read_list:
                url = url + quote_plus(alias) + '&'
            # trim trailing '&'
            url = url[:-1]
        else:
            raise Exception('Tried to read from Exosite, but no alias given')
        return ReadHandler(self.send_get(url, headers))

    def http_long_poll(self, dataport, timeout_ms=None, modify_ts=None):
        '''
            read_list: List of data sources to read from
        '''
        headers = {
            'User-Agent' : self._user_agent,
            'Accept' : self._http_accept,
        }
        if timeout_ms:
            headers['Request-Timeout'] = bytes(timeout_ms)
        headers.update(self._auth_header())
        if modify_ts is not None:
            headers['If-Modified-Since'] = str(int(modify_ts))

        url = self._stack_alias_url() + '?' + quote_plus(dataport)

        return ReadHandler(self.send_get(url, headers))

    def _http_write_dict(self, write_dict):
        """ Wrapper function for http_write.
            url_encodes a python dictionary and calls
            http_write.
            Cannot handle nested dicts. """
        body = urlencode(write_dict) if self._form_encode else write_dict
        resp = self._http_write(body)
        return resp

    def _http_write(self, body):
        """ Method to write to a device dataport. """
        headers = {
            'User-Agent' : self._user_agent,
            'Content-Type' : self._http_content_type,
            'Content-Length' : str(len(body))
        }
        headers.update(self._auth_header())
        url = self._stack_alias_url()
        return WriteHandler(self.send_post(url, headers, body))

    def _http_record(self, record_dict):
        """ Method to record to a device dataport.

            The POST data must look like the following:

            ::

                alias=<alias 1>&<timestamp 1>=<value 1>&<timestamp 2>=<value 2>&alias=<alias 2>&<timestamp 3>=<value 3>&<timestamp 4>=<value 4>

            Since this can't be easily decoded with a Python dictionary, the schema
            for the input dictionary must be as follows:

                {'alias1': {'ts1': 'some data', 'ts2': 'some more data'}, 'alias2': {'ts2': 'data data data', 'ts3': 'datum dateo''}}
        """
        data = ''
        record_dict_len = len(record_dict)
        record_keys = list(record_dict)
        for i in range(0, record_dict_len):
            if i > 0:
                data += '&'
            data += 'alias={}'.format(record_keys[i])
            ts_dict = record_dict[record_keys[i]]
            ts_dict_len = len(ts_dict)
            ts_keys = list(ts_dict)
            for j in range(0, ts_dict_len):
                data += '&{}={}'.format(ts_keys[j], ts_dict[ts_keys[j]])


        LOG.debug("Recording data: {}".format(data))
        headers = {
            'User-Agent' : self._user_agent,
            'Content-Type' : self._http_content_type,
            'Content-Length' : str(len(data))
        }
        headers.update(self._auth_header())
        url = self._stack_record_url()
        return RecordHandler(self.send_post(url, headers, data))

    def _http_readwrite(self, read_list, write_dict):
        '''
            read_list: List of data source to read from
            write_dict: Dictionary of k,v to write to.  (Do not url encode)
        '''
        body = urlencode(write_dict) if self._form_encode else write_dict
        headers = {
            'User-Agent' : self._user_agent,
            'Accept' : self._http_accept,
            'Content-Type' : self._http_content_type,
            'Content-Length' : str(len(body))
        }
        headers.update(self._auth_header())
        url = self._stack_alias_url()
        if read_list > 0:
            url = url + '?'
            for alias in read_list:
                url = url + quote_plus(alias) + '&'
            url = url[:-1] # strip off last '&'
        return ReadWriteHandler(self.send_post(url, headers, body))

    def _http_activate(self):
        """ Method to activate device.
            Returns ActivationHandler() object. """

        if not self.using_tls():
            body = 'id={}'.format(self.murano_id())
        else:
            LOG.critical("TLS Client Cert reprovisioning not yet implemented.")
        headers = {
            'User-Agent' : self._user_agent,
            'Content-Type' : self._http_content_type
        }
        _ = headers.update({'Content-Length': str(len(body))}) if not body is None else None
        url = self.murano_host() + 'provision/activate'
        return ActivationHandler(self.send_post(url, headers, body))

    def _log_curl_cmd(self, method, url, headers, body=None):
        curl_headers = ' '.join(
            [' \\\n    -H "{}: {}"'.format(key, value) \
                for (key, value) in headers.items()
            ]
        )
        cmd = "\ncurl -X {}".format(method) + \
              " \\\n    '{}'".format(url) + \
              "{}".format(curl_headers)
        cmd = cmd + " \\\n    -d '{}'".format(body) if not body is None else cmd
        cmd = cmd + " \\\n    --cert {} \\\n    --key {}".format(
            self.certfile(), self.pkeyfile()) if self.using_tls() else cmd
        cmd = cmd + " \\\n    --cacert {}".format(self.murano_cacert()) if self.murano_cacert() else cmd
        LOG.debug(cmd)

    def send_get(self, url, headers, stream=None, timeout=None):
        """ Method wrapper for requests.get().
            Input req_info: a tuple describing the type of request. Used for data usage
            logging and statistics collection.

            There is one special case in this method. If stream=True,
            then it is assumed that the caller is using wants the raw 'requests'
            object so she may iterate over the streaming content for larger
            file downloads from the Murano content area. Otherwise, the return
            object is of type exo.handlers.Requests_Response.
         """
        cert = None
        try:
            if not self._session:
                self._session = requests.Session()

            if self.using_tls():
                cert = (self.certfile(), self.pkeyfile())
                LOG.debug("Using {} auth for Murano Connection..."
                          .format(cert))
            self._log_curl_cmd('GET', url, headers)
            response = self._session.get(
                url,
                headers=headers,
                timeout=self.timeout() if timeout is None else timeout,
                stream=stream,
                verify=self.murano_cacert() or True,
                cert=cert
            )
        except requests.exceptions.ReadTimeout as __e:
            LOG.info("Caught exception {!r}".format(__e))
            if self._raise:
                raise
            # Treat all exceptions as Timeouts.
            # funky way of giving Requests_Response the
            # object members it needs to perform its duties.
            return type('requests_override', (object,),
                        {'status_code': Http_ReadWriteCodes.Timeout,
                         'text': str(__e),
                         'iter_content': None,
                         'headers': {}})

        LOG.debug("Stream option set to {}".format(stream))
        return response

    def send_post(self, url, headers, body):
        """ Method wrapper for requests.post().
            Input req_info: a tuple describing the type of request. Used for data usage
            logging and statistics collection.
        """
        cert = None
        try:
            if not self._session:
                self._session = requests.Session()
            if self.using_tls():
                cert = (self.certfile(), self.pkeyfile())
                LOG.debug("Using {} auth for Murano Connection..."
                          .format(cert))

            self._log_curl_cmd('POST', url, headers, body=body)
            response = self._session.post(
                url,
                data=body,
                headers=headers,
                timeout=self.timeout(),
                verify=self.murano_cacert() or True,
                cert=cert
            )

        except requests.exceptions.ReadTimeout as __e:
            LOG.info("Caught exception {!r}".format(__e))
            if self._raise:
                raise
            # Treat all exceptions as Timeouts.
            # funky way of giving Requests_Response
            # the object members it needs to perform its duties.
            return type('requests_override', (object,),
                        {'status_code': Http_ReadWriteCodes.Timeout,
                         'text': str(__e),
                         'iter_content': None,
                         'headers': {}})

        return response

    def exosite_timestamp(self):
        """
            Gets the exosite server timestamp from m2.exosite.io/timestamp.

            Returns:
                Timestamp as int.
        """
        headers = {
            'User-Agent' : self._user_agent,
            'Accept' : "*/*"
        }
        return self.send_get(self.murano_host() + 'timestamp', headers)

    def http_activate(self):
        """ Instead of over-riding http_activate(), here is
            logic to deal with activating self. """

        if not self.is_activated():
            act_handler = self._http_activate()
            if act_handler.activated:
                LOG.info("Activation succeeded.")
                self.set_murano_token(act_handler.body)
            else:
                LOG.warning(
                    "Activation for {!r} failed: {!r}".format(
                        str(self), str(act_handler))
                )
        else:
            LOG.info("Already activated.")

    def http_write(self, dataport, value):
        """ Member function that returns WriteHandler(). """
        body = {dataport: value} if self._form_encode else '{}={}'.format(dataport, value)
        LOG.debug("writing ({0}): {1}".format(dataport, value))
        headers = {
            'User-Agent' : self._user_agent,
            'Content-Type' : self._http_content_type,
            'Content-Length' : str(len(body))
        }
        headers.update(self._auth_header())
        url = self._stack_alias_url()

        write_hand = WriteHandler(self.send_post(url, headers, body))
        if write_hand.success:
            pass
        elif write_hand.code == Http_ReadWriteCodes.Unauthorized:
            LOG.debug(
                "BAD TOKEN. Writing {!r} to dataport {!r} failed: {!r}".format(
                    body, dataport, str(write_hand))
            )
            self.set_activated(False)
        else:
            LOG.debug("Error when trying to write {!r} to dataport {!r}: {!r}"
                      .format(body, dataport, str(write_hand)))
        self._online = write_hand.online
        return write_hand

    def http_write_multiple(self, write_dict):
        """ Member function that returns WriteHandler(). """

        LOG.debug("Writing {0}".format(write_dict))
        write_hand = self._http_write_dict(write_dict)
        if write_hand.success:
            pass
        elif write_hand.code == Http_ReadWriteCodes.Unauthorized:
            LOG.debug(
                "BAD TOKEN. Writing {!r} failed: {!r}".format(
                    write_dict, str(write_hand))
            )
            self.set_activated(False)
        else:
            LOG.debug("Error when trying to write {!r}: {!r}".format(
                write_dict, str(write_hand))
                     )
        self._online = write_hand.online
        return write_hand

    def http_record(self, record_dict):
        """ Member function that returns WriteHandler(). """
        # body = ''
        # for dataport in record_dict.keys():
        #   body += '{0}={1}&'.format(dataport, record_dict[dataport])
        # body = body[:-1] # strip off last '&'
        LOG.debug("Recording {0}".format(record_dict))

        record_hand = self._http_record(record_dict)
        if record_hand.success:
            pass
        elif record_hand.code == Http_ReadWriteCodes.Unauthorized:
            LOG.debug(
                "BAD TOKEN. Recording {!r} failed: {!r}".format(
                    record_dict, str(record_hand))
            )
            self.set_activated(False)
        else:
            LOG.debug("Error when trying to write {!r}: {!r}"
                      .format(record_dict, str(record_hand)))
        self._online = record_hand.online
        return record_hand

    def http_read(self, read_list):
        """ Member function that utilizes ReadHandler(). """
        if not isinstance(read_list, list):
            read_list = [read_list]
        read_hand = self._http_read(read_list)
        if read_hand.success:
            pass
        elif read_hand.code == Http_ReadWriteCodes.Unauthorized:
            LOG.debug("BAD TOKEN. Reading {!r} failed: {!r}"
                      .format(read_hand, str(read_hand)))
            self.set_activated(False)
        else:
            LOG.debug("Error when trying to read from dataport(s) {}: {}"
                      .format(read_list, read_hand))
        self._online = read_hand.online
        return read_hand

    def http_read_write(self, read_list, write_dict):
        """ Member function that utilizes ReadWriteHandler(). """
        readwrite_hand = self._http_readwrite(read_list, write_dict)
        if readwrite_hand.success:
            pass
        elif readwrite_hand.code == Http_ReadWriteCodes.Unauthorized:
            LOG.debug("BAD TOKEN. Setting activated = False. Unable to read {!r} nor write {!r}: {!r}"
                      .format(read_list, write_dict, str(readwrite_hand)))
            self.set_activated(False)
        else:
            LOG.debug("Error when trying to read from dataport(s) {!r} and write {!r}: {!r}"
                      .format(read_list, write_dict, str(readwrite_hand)))
        self._online = readwrite_hand.online
        return readwrite_hand




# pylint: disable=C0325,C0103
from murano_client.commands.http import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args

class ExoCommand(ExositeConnection):
    """Record <value> to <alias> at time <timestamp>.

  usage:
    record <timestamp> <alias> <data> [options]

  options:
    -h --help                   Show this screen.
    <timestamp>                 The timestamp (int/float) at which to record the data.
    <alias>                     A Murano Product resource.
    <data>                      The data to write or record.

    """

    def execute(self):
        from murano_client.http import MuranoHTTP
        from murano_client.ini import Device

        if self.global_args.get('--file'):
            api = Device(self.global_args.get('--file'))
        else:
            api = MuranoHTTP(**build_api_opts_from_docopt_args(self.global_args))

        resp_handler = api.http_record(
            {self.args.get('<alias>'): {
                self.args.get('<timestamp>'): self.args.get('<data>')}
            }
        )

        print("[{}] {}".format(resp_handler.code, resp_handler.body))



# pylint: disable=C0325,C0103
from murano_client.commands.http import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args
import sys


class ExoCommand(ExositeConnection):
    """Write <value> to <alias> resource.

  usage:
    write <alias> <data> [options]

  options:
    -h --help                   Show this help screen.
    <alias>                     A Murano Product resource.
    <data>                      The data to write or record. If <data> is the '-'
                                character, then data is read from stdin and a payload
                                is generated from each line (i.e. '\\n' delineated).

    """
    Name = 'write'
    def execute(self):
        from murano_client.http import MuranoHTTP
        from murano_client.ini import Device

        if self.global_args.get('--file'):
            api = Device(self.global_args.get('--file'))
        else:
            api_opts = build_api_opts_from_docopt_args(self.global_args)
            api = MuranoHTTP(**api_opts)

        alias = self.args.get('<alias>')
        data = self.args.get('<data>')
        if data == '-':
            sys.stderr.write("Reading from STDIN.\n")
            def read_stdin():
                """ generator for reading stdin. """
                readline = sys.stdin.readline()
                while readline:
                    yield readline
                    readline = sys.stdin.readline()
            for line in read_stdin():
                payload = line.strip()
                resp_handler = api.http_write(alias, payload)
                print("[{}] {}".format(resp_handler.code, resp_handler.body))

        else:
            resp_handler = api.http_write(alias, data)
            print("[{}] {}".format(resp_handler.code, resp_handler.body))



# pylint: disable=C0325,C0103,C0111,W1202,C0301
from __future__ import print_function
import logging
import sys
from murano_client.commands.mqtt import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args

LOG = logging.getLogger('GDC.MQTT.pub')

class ExoCommand(ExositeConnection):
    r"""Publish data to Murano resources using the MQTT protocol.

  usage:
    publish <topic> <value> [options]

  options:
    -h --help                   Show this screen.
    <topic>                     Mqtt topic to publish to Murano resource.
    <value>                     The value to publish. If <value> is -, publish
                                data line-by-line by piping data via stdin.
    -Q --qos <qos>              Determine the Quality of Service for published
                                messages (default=1). Possible values:

                                    0 - Message sent At Most Once (supported)
                                    1 - Message sent At Least Once (supported)
                                    2 - Message sent Exactly Once (not supported)

  examples:
    # Batch Record Example
    $ gdc --debug DEBUG \
         --host t41hp23nod8s00000.m2.exosite.io \
         --cacert Murano_Root_CA.cer \
         -C device-0001.pem \
         -K device-0001-key.pem \
         mqtt \
         publish \
         \$resource.batch \
         '[{"timestamp":1526046488, "values": {"test": "some test data"}}, {"timestamp":1526046488, "values": {"temperature": 98.7}}]'

    # Simple Publish Example
    $ gdc --pid f5330e5s8cho0000 \
         --cacert Murano_Root_CA.cer \
         -C device-0001.pem \
         -K device-0001-key.pem \
         mqtt \
         publish \
         \$resource/temperature \
         98.7
    """
    Name = 'publish'
    def execute(self):
        from murano_client.ini import Device
        from murano_client.mqtt import MuranoMQTT, MqttPublishHandler
        import time

        Qos = 1
        if self.global_args.get('--qos'):
            Qos = int(self.global_args.get('--qos'))

        if self.global_args.get('--file'):
            mqtt_client = Device(self.global_args.get('--file'))

        else:
            api_opts = build_api_opts_from_docopt_args(self.global_args)
            mqtt_client = MuranoMQTT(**api_opts)

        def on_message(client, userdata, msg):
            """ Override default on_message function. """
            LOG.debug("recieved message: {}".format(msg))

        mqtt_client.on_message = on_message
        mqtt_client.start()
        mqtt_client.loop_start()

        topic = self.args.get('<topic>')
        if self.args.get('<value>') == '-':
            LOG.debug("Reading from STDIN.")
            def read_stdin():
                readline = sys.stdin.readline()
                while readline:
                    yield readline
                    readline = sys.stdin.readline()
            for line in read_stdin():
                payload = line.strip()
                LOG.debug("topic={}, qos={}, payload={}".format(topic, Qos, payload))
                before = time.time()
                handler = MqttPublishHandler(mqtt_client.publish(
                    topic,
                    payload,
                    qos=Qos
                ))
                handler.wait_for_publish()
                sys.stderr.write('{}\n'.format(time.time()-before))
        else:
            payload = self.args.get('<value>')
            LOG.debug("topic={}, qos={}, payload={}".format(topic, Qos, payload))
            handler = MqttPublishHandler(mqtt_client.publish(
                topic,
                payload,
                qos=Qos
            ))
            handler.wait_for_publish()
            sys.exit(0 if handler.is_published() else 1)



# pylint: disable=C0325,C0103
from murano_client.commands.http import Command as ExositeConnection
from murano_client.commands import build_api_opts_from_docopt_args

class ExoCommand(ExositeConnection):
    """Long poll <alias> for <request-timeout> since <if-modified-since>.

  usage:
    poll <alias> <request-timeout> [<if-modified-since>] [options]

  options:
    -h --help                   Show this screen.
    <alias>                     A Murano Product resource.
    <request-timeout>           The amount of time, in milliseconds, to long-poll on <alias>.
    <if-modified-since>         The timestamp (unix epoch) from which to check for new data on <alias>.


    """
    Name = 'poll'
    def execute(self):
        from murano_client.http import MuranoHTTP
        from murano_client.ini import Device

        if self.global_args.get('--file'):
            api = Device(self.global_args.get('--file'))
        else:
            api_opts = build_api_opts_from_docopt_args(self.global_args)
            api = MuranoHTTP(**api_opts)

        alias = self.args.get('<alias>')
        resp_handler = api.http_long_poll(
            alias,
            self.args.get('<request-timeout>'),
            self.args.get('<if-modified-since>') if self.args.get('<if-modified-since>') else None
        )

        print("[{}] {}".format(resp_handler.code, resp_handler.body))



"""
    This module contains the MuranoClient class that can be used to
    quickly create an object that can use either HTTP or MQTT.
"""
# pylint: disable=W1202

import time
import threading
import math
import logging
import sys
import six
import murano_client.constants
from murano_client.http import MuranoHTTP
from murano_client.mqtt import MuranoMQTT
if six.PY2:
    import Queue as queue
    from urlparse import urlparse
elif six.PY3:
    import queue
    from urllib.parse import urlparse


LOG = logging.getLogger(name='MURANO_CLIENT.Client')

class StoppableThread(threading.Thread):
    """Thread class with a stop() method. The thread itself has to check
    regularly for the stopped() condition."""

    def __init__(self, **kwargs):
        LOG.info("starting new thread with kwargs: {}".format(kwargs))
        super(StoppableThread, self).__init__(**kwargs)
        self.__stop_event = threading.Event()
        self.__started = False

    def start(self):
        if not self.__started:
            super(StoppableThread, self).start()
            self.__started = True
        else:
            LOG.debug("Start method called, but already started. Everything OK.")

    def is_started(self):
        """ Method for checking whether thread has been started."""
        return self.__started

    def stop(self):
        """ Raise the stop event."""
        self.__stop_event.set()

    def is_stopped(self):
        """ Check if the stop event is set."""
        return self.__stop_event.is_set()

class WatchQueue(queue.Queue):
    """
        In an effort to stop writing try-except-Empty all over the place
        for handling queue.get() calls, implimenting a queue that returns
        None if the queue is empty.
    """
    def __init__(self, **kwargs):
        LOG.debug("kwargs: {}".format(kwargs))
        queue.Queue.__init__(self)

    def safe_get(self, timeout=None):
        """
            This function will block forever unless timeout is overridden.
            Returns data from the queue or None if timeout is reached with
            no data.
        """
        try:
            return self.get(timeout=timeout)
        except queue.Empty:
            return None

class InboundPayload(object):
    def __init__(self, **kwargs):
        self.payload = None
        self.timestamp = None
        self.resource = None
        protocol = kwargs.get('outbound_protocol')
        if protocol == 'mqtt':
            self.payload = kwargs.get('message').payload
            self.timestamp = kwargs.get('message').topic.split('/')[2]
            self.resource = kwargs.get('message').topic.split('/')[1]
        elif protocol == 'https':
            self.payload = kwargs.get('message').split('=')[1]
            self.timestamp = time.time()
            self.resource = kwargs.get('message').split('=')[0]

    def __str__(self):
        return "{}:{}={}".format(self.resource, self.timestamp, self.payload)

class OutboundPayload(object):
    def __init__(self, **kwargs):
        self.payload = kwargs.get('payload')
        self.timestamp = kwargs.get('timestamp')

        if kwargs.get('outbound_protocol') == 'mqtt':
            self.resource = '$resource/{}'.format(kwargs.get('resource'))
        elif kwargs.get('outbound_protocol') == 'https':
            self.resource = kwargs.get('resource')

    def __str__(self):
        return "{}:{}={}".format(self.resource, self.timestamp, self.payload)


class HTTPWatchThread(StoppableThread):
    def __init__(self, **kwargs):
        _ = LOG.setLevel(getattr(logging, kwargs.get('debug'))) if kwargs.get('debug') else None
        LOG.debug("kwargs: {}".format(kwargs))
        super(HTTPWatchThread, self).__init__()
        self.setDaemon(True)
        self.watch = kwargs.get('watch_function')
        self.name = '-'.join([self.__class__.__name__, kwargs.get('resource')])
        self.resource = kwargs.get('resource')
        self.watch_timeout = kwargs.get('http_timeout')
        self.inbound_queue = kwargs.get('inbound_queue')

    def run(self):
        since = 0.0

        while not self.is_stopped():
            LOG.info("(re)starting watch of: {}".format(self.resource))

            if since == 0.0:
                response = self.watch(self.resource)
            else:
                response = self.watch(
                    self.resource,
                    self.watch_timeout,
                    modify_ts=since)
            now = time.time()
            tdiff = now - since
            # TODO: the readhandler needs to parse a human-readable last-modified header.
            # since = response.last_modified or math.ceil(now)
            since = math.ceil(now)
            LOG.info('{} watched for {}s, got: {}'.format(self.__class__.__name__, tdiff, response))
            if response.code in [304, 204]:
                LOG.info("No new watch data.")
                continue
            if response.success:
                self.inbound_queue.put(
                    InboundPayload(
                        message=response.body,
                        outbound_protocol='https'
                    )
                )
            else:
                LOG.warning("http watcher was not successful: {}".format(response))


class HTTPTellThread(StoppableThread):
    """
        needs an outbound_queue
        objects going into the outbound queue need
        to be instances of OutboundPayload
    """
    def __init__(self, **kwargs):
        _ = LOG.setLevel(getattr(logging, kwargs.get('debug'))) if kwargs.get('debug') else None
        LOG.debug("kwargs: {}".format(kwargs))
        super(HTTPTellThread, self).__init__()
        self.setDaemon(True)
        self.name = self.__class__.__name__
        self.tell = kwargs.get('tell_function')
        self.outbound_queue = kwargs.get('outbound_queue')

    def run(self):
        LOG.info("Starting tell thread.")
        while not self.is_stopped():
            tell = self.outbound_queue.safe_get(timeout=1.0)
            if tell:
                LOG.debug("Sending tell data: {}".format(tell))
                response = self.tell(
                    {tell.resource: {tell.timestamp: tell.payload}})
                LOG.info("Tell response: {}".format(response))
        LOG.critical("{} has stopped!".format(self))

class MQTTTellThread(StoppableThread):
    """
        needs an outbound_queue
        objects going into the outbound queue need
        to be instances of OutboundPayload
    """
    def __init__(self, **kwargs):
        _ = LOG.setLevel(getattr(logging, kwargs.get('debug'))) if kwargs.get('debug') else None
        LOG.debug("kwargs: {}".format(kwargs))
        super(MQTTTellThread, self).__init__()
        self.setDaemon(True)
        self.name = self.__class__.__name__
        self.tell = kwargs.get('tell_function')
        self.outbound_queue = kwargs.get('outbound_queue')

    def run(self):
        LOG.info("starting tell thread run")
        while not self.is_stopped():
            LOG.debug("back to top of tell-thread main loop")
            tell = self.outbound_queue.safe_get(timeout=1.0)
            if tell and tell.resource:
                # TODO: when either or both of MUR-6791 and MUR-7122 are fixed, the commented
                # code, below can be utilized to do records. For now, 
                # payload = [{'timestamp': tell.timestamp, 'values': {tell.resource: tell.payload}}]
                # LOG.info("telling to resource {}: {}".format(tell.resource, payload))
                LOG.info("telling to resource {}: {}".format(tell.resource, tell.payload))
                # response = self.tell('$resource.batch', json.dumps(payload), qos=1)
                response = self.tell(
                    tell.resource,
                    # json.dumps(payload),
                    tell.payload,
                    qos=1)
                LOG.info(response)
            else:
                LOG.debug(
                    "No resource in outbound payload object. Cannot tell Murano anything: {}"
                    .format(tell))
        LOG.critical("{} has stopped!".format(self))


class MQTTWatchThread(StoppableThread):
    def __init__(self, **kwargs):
        _ = LOG.setLevel(getattr(logging, kwargs.get('debug'))) if kwargs.get('debug') else None
        LOG.debug("kwargs: {}".format(kwargs))
        super(MQTTWatchThread, self).__init__()
        self.setDaemon(True)
        self.watch = kwargs.get('watch_function')
        self.name = '-'.join([self.__class__.__name__, kwargs.get('resource')])
        self.resource = kwargs.get('resource')
        self.watch_timeout = kwargs.get('http_timeout')
        self.inbound_queue = kwargs.get('inbound_queue')

    def run(self):
        since = 0.0
        while not self.is_stopped():
            LOG.info("starting watch of: {}".format(self.resource))

            message = self.watch(timeout=self.watch_timeout)
            now = time.time()
            tdiff = now - since
            since = math.ceil(now)
            LOG.info('watch time {}: {}'.format(tdiff, message))
            if message:
                LOG.info(
                    "putting new message in inbound queue!")
                self.inbound_queue.put(InboundPayload(message=message, outbound_protocol='mqtt'))
        LOG.critical("{} has stopped!".format(self))

class MuranoClientException(Exception):
    pass

class MuranoClient(MuranoHTTP, MuranoMQTT):
    def __init__(self, **kwargs):
        """
            The inbound queue is designed so that objects of this class can call
            inbound_queue.safe_get([timeout]) and the payload objects from Murano
            will be formatted in a predictable way.

                Payload format of items in inbound_queue:

                    <resource>=
        """
        LOG.debug("Initializing MuranoHTTP Client...")
        self.inbound_queue = WatchQueue() # data from murano
        self.outbound_queue = WatchQueue() # data to murano
        self.outbound_protocol = None

        murano_host = kwargs.get('murano_host', None)
        if murano_host:
            self.outbound_protocol = urlparse(murano_host).scheme

        LOG.info("Using protocol {}".format(self.outbound_protocol))

        if not self.outbound_protocol:
            if six.PY3:
                sys.tracebacklimit = None
            else:
                sys.tracebacklimit = 0
            raise MuranoClientException(
                "Cannot parse supported protocol from murano_host parameter: {}"
                .format(kwargs.get('murano_host')))

        if not kwargs.get('watchlist'):
            if six.PY3:
                sys.tracebacklimit = None
            else:
                sys.tracebacklimit = 0
            raise MuranoClientException(
                "Required parameter 'watchlist' empty or not provided: {}"
                .format(kwargs.get('watchlist')))

        if self.outbound_protocol == 'mqtt': # no need for this for http
            kwargs['queue_inbound_messages'] = WatchQueue()
            MuranoMQTT.__init__(self, **kwargs)
        elif self.outbound_protocol == 'https':
            MuranoHTTP.__init__(self, **kwargs)

        if not self.murano_host():
            raise MuranoClientException("no murano_host specified.")

        self.watch_threads = []

        # ###################################################### #
        # ########### HTTP HTTP HTTP HTTP HTTP ################# #
        # ###################################################### #

        if self.outbound_protocol == 'https':
            for resource in kwargs.get('watchlist'):
                LOG.info('watching: {}'.format(resource))
                self.watch_threads.append(
                    HTTPWatchThread(
                        watch_function=self.http_long_poll,
                        inbound_queue=self.inbound_queue,
                        resource=resource,
                        http_timeout=kwargs.get(
                            'http_timeout',
                            murano_client.constants.DEFAULT_HTTP_TIMEOUT)*1000
                    )
                )
            LOG.critical("done init'ing thread objects.")

            self.tell_thread = HTTPTellThread(
                outbound_queue=self.outbound_queue,
                tell_function=self.http_record)

        # ###################################################### #
        # ########### MQTT MQTT MQTT MQTT MQTT ################# #
        # ###################################################### #

        elif self.outbound_protocol == 'mqtt':
            for resource in kwargs.get('watchlist'):
                LOG.info('watching: {}'.format(resource))
                self.watch_threads.append(
                    MQTTWatchThread(
                        watch_function=kwargs['queue_inbound_messages'].safe_get,
                        inbound_queue=self.inbound_queue,
                        resource=resource,
                        http_timeout=kwargs.get('http_timeout')*1000 if kwargs.get('http_timeout') else 10*1000
                    )
                )
            LOG.info("done init'ing thread objects.")
            # self.start_watchers()
            # LOG.info(self.watch_threads)

            self.tell_thread = MQTTTellThread(
                outbound_queue=self.outbound_queue,
                tell_function=self.publish)

            # tell_thread.start()
            # LOG.info("tell thread started")
        self.monitor_thread = StoppableThread(
            name='Monitor',
            target=self._monitor_thread,
            args=())
        self.monitor_thread.setDaemon(True)

    def stop_all(self):

        LOG.critical("Stopping all threads!!!")
        self.tell_thread.stop()
        LOG.critical("Waiting for {} to stop...".format(self.tell_thread))
        self.tell_thread.join()
        for watcher in self.watch_threads:
            if not watcher.is_stopped():
                watcher.stop()
            LOG.critical("Waiting for {} to stop...".format(watcher))
            watcher.join(timeout=0.25)
        if self.outbound_protocol == 'https':
            LOG.critical("Closing HTTP connection!!!")
            self.close_session()
        if self.outbound_protocol == 'mqtt':
            LOG.critical("Closing MQTT connection!!!")
            def on_disconnect_override(client, userdata, rc): # pylint: disable=C0103
                """ MQTT on_disconnect handler override function."""
                LOG.critical("Default reconnect behavior overridden.")
                LOG.critical("Disconnecting: client: {} userdata: {} rc: {}"
                             .format(client, userdata, rc))
            self.on_disconnect = on_disconnect_override
            self.disconnect()

        self.monitor_thread.stop()
        LOG.critical("Waiting for {} to stop...".format(self.monitor_thread))
        self.monitor_thread.join()
        LOG.critical("All threads have stopped.")
        return

    def start_client(self):
        if self.outbound_protocol == 'mqtt':
            self.start()
            self.loop_start()
        self.start_watchers()
        self.tell_thread.start()
        self.monitor_thread.start()

    def _monitor_thread(self):
        while not self.monitor_thread.is_stopped():
            for thread in threading.enumerate():
                LOG.info("running thread: {}".format(thread))
            LOG.info("inbound_queue size: {}".format(self.inbound_queue.qsize()))
            LOG.info("outbound_queue size: {}".format(self.outbound_queue.qsize()))
            time.sleep(5)
        LOG.critical("{} has stopped!".format(self.monitor_thread))

    def start_watchers(self):
        LOG.info("Starting watchers.")
        for thread in self.watch_threads:
            LOG.debug("starting: {}".format(thread))
            thread.setDaemon(True)
            thread.start()

    def tell(self, **kwargs):
        o_p = OutboundPayload(
            resource=kwargs.get('resource'),
            timestamp=kwargs.get('timestamp'),
            payload=kwargs.get('payload'),
            outbound_protocol=self.outbound_protocol)
        LOG.info("telling to outbound queue: {}".format(o_p))
        self.outbound_queue.put(o_p)

    def watch(self, **kwargs):
        if kwargs.get('timeout'):
            from_murano = self.inbound_queue.safe_get(timeout=kwargs.get('timeout'))
            LOG.info("got from murano, putting in inbound queue: {}".format(from_murano))
            return from_murano
        else:
            while True:
                from_murano = self.inbound_queue.safe_get(timeout=1.0)
                if from_murano:
                    LOG.info("got from murano, putting in inbound queue: {}".format(from_murano))
                    return from_murano
                else:
                    LOG.debug("No data from murano: {}".format(from_murano))


    def client_activate(self):
        if self.outbound_protocol == 'mqtt':
            self.start(activate_only=True)
        elif self.outbound_protocol == 'https':
            self.http_activate()
